import numpy as np

from pypop7.optimizers.rs.rs import RS


class CSA(RS):
    """Simulated Annealing (SA) designed by Corana et al., ACM-TOMS, 1987.

    Reference
    ---------
    Kirkpatrick, S., Gelatt, C.D. and Vecchi, M.P., 1983.
    Optimization by simulated annealing.
    Science, 220(4598), pp.671-680.
    https://science.sciencemag.org/content/220/4598/671

    Corana, A., Marchesi, M., Martini, C. and Ridella, S., 1987.
    Minimizing multimodal functions of continuous variables with the “simulated annealing” algorithm.
    ACM Transactions on Mathematical Software, 13(3), pp.262-280.
    https://dl.acm.org/doi/abs/10.1145/29380.29864
    https://dl.acm.org/doi/10.1145/66888.356281

    https://esa.github.io/pygmo2/algorithms.html#pygmo.simulated_annealing
    """
    def __init__(self, problem, options):
        RS.__init__(self, problem, options)
        step_vector = (self.initial_upper_boundary - self.initial_lower_boundary) * 0.01
        self.v = options.get('v', step_vector)  # starting step vector
        self.T = options.get('T', 10.0)  # starting temperature
        self.N_S = options.get('N_S', 10)  # for step variation
        self.c = options.get('c', 2 * np.ones(self.ndim_problem,))  # for step variation criterion
        self.N_T = options.get('N_T', 10)  # for temperature reduction
        self.r_T = options.get('r_T', 0.85)  # for temperature reduction coefficient
        self.n = np.zeros((self.ndim_problem,))  # for step variation
        self.parent_x = np.copy(self.best_so_far_x)
        self.parent_y = np.copy(self.best_so_far_y)
        self.n_T_adjustments = 0  # number of temperature adjustments

    def initialize(self, args=None):
        if self.x is None:  # starting point
            x = self.rng_initialization.uniform(self.initial_lower_boundary,
                                                self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        y = self._evaluate_fitness(x, args)
        self.parent_x, self.parent_y = x, y
        return y

    def iterate(self, args=None):  # perform a cycle of random moves
        fitness = []
        for h in np.arange(self.ndim_problem):
            x = np.copy(self.parent_x)
            search_range = (np.maximum(self.parent_x[h] - self.v[h], self.lower_boundary[h]),
                            np.minimum(self.parent_x[h] + self.v[h], self.upper_boundary[h]))
            x[h] = self.rng_optimization.uniform(search_range[0], search_range[1])
            y = self._evaluate_fitness(x, args)
            if self.record_fitness:
                fitness.append(y)
            self._print_verbose_info()
            diff = self.parent_y - y
            if (diff >= 0) or (self.rng_optimization.random() < np.exp(diff / self.T)):
                self.parent_x, self.parent_y = x, y
                self.n[h] += 1
            if self._check_terminations():
                break
        return fitness

    def adjust_step_vector(self):
        for u in range(self.ndim_problem):
            if self.n[u] > 0.6 * self.N_S:
                self.v[u] *= 1 + self.c[u] * (self.n[u] / self.N_S - 0.6) / 0.4
            elif self.n[u] < 0.4 * self.N_S:
                self.v[u] /= 1 + self.c[u] * (0.4 - self.n[u] / self.N_S) / 0.4
        self.n = np.zeros((self.ndim_problem,))

    def _collect_results(self, fitness):
        results = RS._collect_results(self, fitness)
        results['n_T_adjustments'] = self.n_T_adjustments
        results['v'] = np.copy(self.v)
        return results

    def optimize(self, fitness_function=None, args=None):
        super(RS, self).optimize(fitness_function)
        fitness = [self.initialize(args)]  # store all fitness generated during search
        while not self._check_terminations():
            for m in range(self.N_T):
                for j in range(self.N_S):
                    fitness.extend(self.iterate(args))
                    if self._check_terminations():
                        break
                if self._check_terminations():
                    break
                self.adjust_step_vector()
            self.T *= self.r_T  # reduce temperature
            self.n_T_adjustments += 1
            self.parent_x, self.parent_y = self.best_so_far_x, self.best_so_far_y
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class RS(Optimizer):
    """Random (stochastic) Search (optimization) (RS).

    This is the **base** class for all RS classes (also including **local search**, particularly its
    randomized versions).

    .. note:: At least its two methods (`initialize`, `iterate`) should be implemented by its subclass.

       `The topic of local search was reinvigorated in the early 1990s by surprisingly good results for
       large ... problems ... and by the incorporation of randomness, multiple simultaneous searches,
       and other improvements. <>`_

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `1000`);
              and with one particular setting (`key`):
                * 'x'     - initial (starting) point (`array_like`).

    Attributes
    ----------
    x     : `array_like`
            initial (starting) point.

    Methods
    -------

    References
    ----------
    Bergstra, J. and Bengio, Y., 2012.
    Random search for hyper-parameter optimization.
    Journal of Machine Learning Research, 13(2).
    https://www.jmlr.org/papers/v13/bergstra12a.html

    Appel, M.J., Labarre, R. and Radulovic, D., 2004.
    On accelerated random search.
    SIAM Journal on Optimization, 14(3), pp.708-731.
    https://epubs.siam.org/doi/abs/10.1137/S105262340240063X

    Solis, F.J. and Wets, R.J.B., 1981.
    Minimization by random search techniques.
    Mathematics of operations research, 6(1), pp.19-30.
    https://pubsonline.informs.org/doi/abs/10.1287/moor.6.1.19

    Schrack, G. and Choit, M., 1976.
    Optimized relative step size random searches.
    Mathematical Programming, 10(1), pp.230-244.
    https://link.springer.com/article/10.1007/BF01580669

    Schumer, M.A. and Steiglitz, K., 1968.
    Adaptive step size random search.
    IEEE Transactions on Automatic Control, 13(3), pp.270-276.
    https://ieeexplore.ieee.org/abstract/document/1098903

    Matyas, J., 1965.
    Random optimization.
    Automation and Remote control, 26(2), pp.246-253.
    https://tinyurl.com/25339c4x  (in Russian)

    Rastrigin, L.A., 1963.
    The convergence of the random search method in the extremal control of a many parameter system.
    Automaton & Remote Control, 24, pp.1337-1342.
    https://tinyurl.com/djfdnpx4

    Brooks, S.H., 1958.
    A discussion of random methods for seeking maxima.
    Operations Research, 6(2), pp.244-251.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.6.2.244
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self.x = options.get('x')  # initial (starting) point
        # reset its default value from 10 to 1000 (since typically more iterations are run)
        if options.get('verbose_frequency') is None:
            self.verbose_frequency = 1000
        self._n_generations = 0  # number of generations

    def initialize(self):
        raise NotImplementedError

    def iterate(self):  # for each iteration (generation)
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def optimize(self, fitness_function=None, args=None):  # for all iterations (generations)
        fitness = Optimizer.optimize(self, fitness_function)
        is_initialization = True
        while not self._check_terminations():
            if is_initialization:
                x = self.initialize()
                is_initialization = False
            else:
                x = self.iterate()
            y = self._evaluate_fitness(x, args)
            if self.record_fitness:
                fitness.append(y)
            self._print_verbose_info(y)
            self._n_generations += 1
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.es.es import ES


class MAES(ES):
    """Matrix Adaptation Evolution Strategy (MAES, (μ/μ_w,λ)-MA-ES).

    Reference
    ---------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700).
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115

    See the official Matlab version from Beyer:
    https://homepages.fhv.at/hgb/downloads/ForDistributionFastMAES.tar
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        self.c_s = None  # for M10 in Fig. 3
        self.alpha_cov = 2  # for M11 in Fig. 3 (α_cov)
        self.c_1 = None  # for M11 in Fig. 3
        self.c_w = None  # for M11 in Fig. 3 (c_μ)
        self.d_sigma = None  # for M12 in Fig. 3 (d_σ)
        self._s_1 = None  # for M10 in Fig. 3
        self._s_2 = None  # for M10 in Fig. 3
        self._fast_version = options.get('_fast_version', False)
        if not self._fast_version:
            self._diag_one = np.diag(np.ones((self.ndim_problem,)))  # for M11 in Fig. 3

    def _set_c_w(self):
        return np.minimum(1 - self.c_1, self.alpha_cov * (self._mu_eff + 1 / self._mu_eff - 2) /
                          (np.power(self.ndim_problem + 2, 2) + self.alpha_cov * self._mu_eff / 2))

    def _set_d_sigma(self):
        return 1 + self.c_s + 2 * np.maximum(0, np.sqrt((self._mu_eff - 1) / (self.ndim_problem + 1)) - 1)

    def initialize(self, is_restart=False):  # for M1 in Fig. 3
        self.c_s = self.options.get('c_s', (self._mu_eff + 2) / (self._mu_eff + self.ndim_problem + 5))
        self.c_1 = self.options.get('c_1', self.alpha_cov / (np.power(self.ndim_problem + 1.3, 2) + self._mu_eff))
        self.c_w = self.options.get('c_w', self._set_c_w())
        self.d_sigma = self.options.get('d_sigma', self._set_d_sigma())
        self._s_1 = 1 - self.c_s
        self._s_2 = np.sqrt(self._mu_eff * self.c_s * (2 - self.c_s))
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        d = np.empty((self.n_individuals, self.ndim_problem))  # search directions
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path
        tm = np.diag(np.ones((self.ndim_problem,)))  # transformation matrix M
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return z, d, mean, s, tm, y

    def iterate(self, z=None, d=None, mean=None, tm=None, y=None, args=None):
        for k in range(self.n_individuals):  # for M3 in Fig. 3 (sample offspring population)
            if self._check_terminations():
                return z, d, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))  # for M4 in Fig. 3
            d[k] = np.squeeze(np.dot(tm, z[k][:, np.newaxis]))  # for M5 in Fig. 3
            y[k] = self._evaluate_fitness(mean + self.sigma * d[k], args)  # for M6 in Fig. 3
        return z, d, y

    def _update_distribution(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        order = np.argsort(y)  # for M8 in Fig. 3
        # for M9, M10, M11 in Fig. 3
        d_w, z_w, zz_w = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,)), None
        if not self._fast_version:
            zz_w = np.zeros((self.ndim_problem, self.ndim_problem))  # for M11 in Fig. 3
        for k in range(self.n_parents):
            d_w += self._w[k] * d[order[k]]
            z_w += self._w[k] * z[order[k]]
            if not self._fast_version:
                zz_w += self._w[k] * np.dot(z[order[k]][:, np.newaxis], z[order[k]][np.newaxis, :])
        # update distribution mean (for M9 in Fig. 3)
        mean += self.sigma * d_w
        # update evolution path (s) and transformation matrix (M)
        s = self._s_1 * s + self._s_2 * z_w  # for M10 in Fig. 3
        if not self._fast_version:
            tm_1 = self.c_1 * (np.dot(s[:, np.newaxis], s[np.newaxis, :]) - self._diag_one)
            tm_2 = self.c_w * (zz_w - self._diag_one)
            tm += 0.5 * np.dot(tm, tm_1 + tm_2)  # for M11 in Fig. 3
        else:
            tm = (1 - 0.5 * (self.c_1 + self.c_w)) * tm + (0.5 * self.c_1) * np.dot(
                np.dot(tm, s[:, np.newaxis]), s[np.newaxis, :])
            for k in range(self.n_parents):
                tm += (0.5 * self.c_w) * self._w[k] * np.dot(d[order[k]][:, np.newaxis], z[order[k]][np.newaxis, :])
        # update global step-size (for M12 in Fig. 3)
        self.sigma *= np.exp(self.c_s / self.d_sigma * (np.linalg.norm(s) / self._e_chi - 1))
        return mean, s, tm

    def restart_initialize(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            z, d, mean, s, tm, y = self.initialize(is_restart)
        return z, d, mean, s, tm, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, d, mean, s, tm, y = self.initialize()
        while True:
            z, d, y = self.iterate(z, d, mean, tm, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, s, tm = self._update_distribution(z, d, mean, s, tm, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                z, d, mean, s, tm, y = self.restart_initialize(z, d, mean, s, tm, y)
        results = self._collect_results(fitness, mean)
        results['s'] = s
        return results



"""Repeat Fig. 3 (a) and (c) from the following paper:
    He, X., Zheng, Z. and Zhou, Y., 2021.
    MMES: Mixture model-based evolution strategy for large-scale optimization.
    IEEE Transactions on Evolutionary Computation, 25(2), pp.320-333.
    https://ieeexplore.ieee.org/abstract/document/9244595
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from optimizers.es.mmes import MMES

sns.set_theme(style='darkgrid')


if __name__ == '__main__':
    # plot Fig. 3 (a) ellipsoid and (c) rosenbrock
    from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock
    ndim_problem = 1000
    for f in [ellipsoid, rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -5 * np.ones((ndim_problem,)),
                   'upper_boundary': 5 * np.ones((ndim_problem,))}
        options = {'fitness_threshold': 1e-8,
                   'max_function_evaluations': 1e8,
                   'seed_rng': 2022,  # not given in the original paper
                   'sigma': 3,
                   'verbose_frequency': 20000,
                   'record_fitness': True,
                   'record_fitness_frequency': 1}
        mmes = MMES(problem, options)
        results = mmes.optimize()
        plt.figure()
        plt.xscale('log')
        plt.yscale('log')
        plt.yticks(ticks=[1e-5, 1e0, 1e5])
        plt.xticks(ticks=[1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
        plt.xlim([1e2, 1e8])
        fitness = results['fitness']
        plt.plot(fitness[:, 0], fitness[:, 1], 'r-')
        plt.xlabel('FEs')
        plt.ylabel('Objective Value')
        plt.show()



"""Repeat the following paper for `EMNA`:
    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5
    (See Chapter 8 Experimental Results in Function Optimization with EDAs in Continuous Domain.)

    There seems to be vey small performance gaps between the original paper and our code, which may be
    ignored for benchmarking. Such gaps may attribute to a slight implementation difference (where our
    code does not use *ad-hoc elitist selection* employed in the original paper).

    Based on the above analysis, we argue that the repeatability of `EMNA` can be well-documented
    (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, griewank
from pypop7.optimizers.eda.emna import EMNA


if __name__ == '__main__':
    ndim_problem = 10

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 2000,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 7.674018032114672
    # vs 8.7201 (from the original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 750,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0
    # vs 5.1166e-02 (from the original paper)

    ndim_problem = 50

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 2000,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 56.717810673268225
    # vs 49.7588 (from the original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600 * np.ones((ndim_problem,)),
               'upper_boundary': 600 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 750,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 1.24590323770534
    # vs 8.7673e-06 (from the original paper)



import numpy as np

from pypop7.optimizers.ga.ga import GA


def get_prob(prob, y1, y2):
    if y2 > y1:
        return_prob = prob + 0.1
        if return_prob > 0.95:
            return_prob = 0.95
    else:
        return_prob = prob - 0.1
        if return_prob < 0.05:
            return_prob = 0.05
    return return_prob


class GENITOR(GA):
    """GENITOR algorithm
    Reference
    --------------
    D. Whitley, S. Dominic, R. Das, C. W. Anderson
    Genetic Algorithm Learning for Neurocontrol Problems
    Machine Learning, 13, 259-284(1993)
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.options = options

    def initialize(self):
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            x[i] = self._initialize_x()
            y[i] = self._evaluate_fitness(x[i])
        prob_cross = self.prob_cross * np.ones((self.n_individuals,))
        return x, y, prob_cross

    def iterate(self, x, y, prob_cross):
        order = np.argsort(y)
        rand = np.random.random()
        # do the crossover work
        if rand < prob_cross[order[0]]:
            x[order[-2]], x[order[-1]] = self.crossover(x[order[0]], x[order[1]], "one_point")
            y[order[-2]] = self._evaluate_fitness(x[order[-2]])
            y[order[-1]] = self._evaluate_fitness(x[order[-1]])
            prob_cross[order[-2]] = get_prob(prob_cross[order[0]], y[order[-2]], y[order[0]])
            prob_cross[order[-1]] = get_prob(prob_cross[order[0]], y[order[-1]], y[order[0]])
        # do the mutate work
        else:
            x[order[0]] = self.mutate(x[order[0]])
            y[order[0]] = self._evaluate_fitness(x[order[0]])
        return x, y, prob_cross

    def mutate(self, x):
        for i in range(self.ndim_problem):
            rand = np.random.random()
            if rand < self.prob_mutate:
                x[i] = x[i] + np.random.random() * 20 - 10
        return x

    def optimize(self, fitness_function=None):
        fitness = GA.optimize(self, fitness_function)
        x, y, prob_cross = self.initialize()
        while True:
            x, y, prob_cross = self.iterate(x, y, prob_cross)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self.__collect_results(fitness)
        return results



import numpy as np

from pypop7.optimizers.es.es import ES


class SEPCMAES(ES):
    """Separable Covariance Matrix Adaptation Evolution Strategy (SEPCMAES, sep-CMA-ES).

    Note that only the diagonal elements of covariance matrix are saved explicitly,
    since all off-diagonal elements are not used for sep-CMA-ES.

    Reference
    ---------
    Ros, R. and Hansen, N., 2008, September.
    A simple modification in CMA-ES achieving linear time and space complexity.
    In International Conference on Parallel Problem Solving from Nature (pp. 296-305).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-540-87700-4_30
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_c = options.get('c_c', 4 / (self.ndim_problem + 4))
        self.options = options
        self.c_s = None
        self.c_cov = None
        self.d_sigma = None
        self._s_1 = None
        self._s_2 = None

    def _set_c_cov(self):
        c_cov = (1 / self._mu_eff) * (2 / np.power(self.ndim_problem + np.sqrt(2), 2)) + (
            (1 - 1 / self._mu_eff) * np.minimum(1, (2 * self._mu_eff - 1) / (
                np.power(self.ndim_problem + 2, 2) + self._mu_eff)))
        c_cov *= (self.ndim_problem + 2) / 3  # for faster adaptation
        return c_cov

    def _set_d_sigma(self):
        d_sigma = np.maximum((self._mu_eff - 1) / (self.ndim_problem + 1) - 1, 0)
        return 1 + self.c_s + 2 * np.sqrt(d_sigma)

    def initialize(self, is_restart=False):
        self.c_s = self.options.get('c_s', (self._mu_eff + 2) / (self.ndim_problem + self._mu_eff + 3))
        self.c_cov = self.options.get('c_cov', self._set_c_cov())
        self.d_sigma = self.options.get('d_sigma', self._set_d_sigma())
        self._s_1 = 1 - self.c_s
        self._s_2 = np.sqrt(self._mu_eff * self.c_s * (2 - self.c_s))
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path for CSA
        p = np.zeros((self.ndim_problem,))  # evolution path for CMA
        c = np.ones((self.ndim_problem,))  # diagonal elements for covariance matrix
        d = np.ones((self.ndim_problem,))  # diagonal elements for covariance matrix
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return z, x, mean, s, p, c, d, y

    def iterate(self, z=None, x=None, mean=None, d=None, y=None, args=None):
        for k in range(self.n_individuals):  # Line 2 of Algorithm 1
            if self._check_terminations():
                return z, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            x[k] = mean + self.sigma * d * z[k]  # Line 3 of Algorithm 1
            y[k] = self._evaluate_fitness(x[k], args)
        return z, x, y

    def _update_distribution(self, z=None, x=None, s=None, p=None, c=None, d=None, y=None):
        order = np.argsort(y)
        zeros = np.zeros((self.ndim_problem,))
        z_w, mean, dz_w = np.copy(zeros), np.copy(zeros), np.copy(zeros)
        for k in range(self.n_parents):  # Line 4, 5 of Algorithm 1
            z_w += self._w[k] * z[order[k]]
            mean += self._w[k] * x[order[k]]  # update distribution mean
            dz = d * z[order[k]]
            dz_w += self._w[k] * dz * dz
        s = self._s_1 * s + self._s_2 * z_w  # Line 6 of Algorithm 1
        # Line 7 of Algorithm 1
        if (np.linalg.norm(s) / np.sqrt(1 - np.power(1 - self.c_s, 2 * self._n_generations))) < (
                (1.4 + 2 / (self.ndim_problem + 1)) * self._e_chi):
            h = np.sqrt(self.c_c * (2 - self.c_c)) * np.sqrt(self._mu_eff) * d * z_w
        else:
            h = 0
        p = (1 - self.c_c) * p + h  # Line 8 of Algorithm 1
        c = (1 - self.c_cov) * c + (1 / self._mu_eff) * self.c_cov * p * p + (
                self.c_cov * (1 - 1 / self._mu_eff) * dz_w)  # Line 9 of Algorithm 1
        # Line 10 of Algorithm 1
        self.sigma *= np.exp(self.c_s / self.d_sigma * (np.linalg.norm(s) / self._e_chi - 1))
        d = np.sqrt(c)  # Line 11 of Algorithm 1
        return mean, s, p, c, d

    def restart_initialize(self, z=None, x=None, mean=None, s=None, p=None, c=None, d=None, y=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            z, x, mean, s, p, c, d, y = self.initialize(is_restart)
            self._n_generations = 0
        return z, x, mean, s, p, c, d, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, x, mean, s, p, c, d, y = self.initialize()
        while True:
            self._n_generations += 1  # Line 1 of Algorithm 1
            z, x, y = self.iterate(z, x, mean, d, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, s, p, c, d = self._update_distribution(z, x, s, p, c, d, y)
            self._print_verbose_info(y)
            if self.is_restart:
                z, x, mean, s, p, c, d, y = self.restart_initialize(z, x, mean, s, p, c, d, y)
        results = self._collect_results(fitness, mean)
        results['s'] = s
        results['p'] = p
        results['d'] = d
        return results



import os
import numpy as np

from pypop7.benchmarks import base_functions
from pypop7.benchmarks.base_functions import squeeze_and_check, BaseFunction


# helper functions
def generate_rotation_matrix(func, ndim, seed):
    """Generate a random rotation matrix of dimension [`ndim` * `ndim`], sampled normally.

        Note that the generated rotation matrix will be automatically stored in txt form for further use.

    :param func: function name, a `str` or `function` object.
    :param ndim: number of dimensions of the rotation matrix, an `int` scalar.
    :param seed: seed for random number generator, a `int` scalar.
    :return: rotation matrix, a [`ndim` * `ndim`] ndarray.
    """
    if hasattr(func, '__call__'):
        func = func.__name__
    data_folder = 'pypop_benchmarks_input_data'
    if not os.path.exists(data_folder):
        os.mkdir(data_folder)
    data_path = os.path.join(data_folder, 'rotation_matrix_' + func + '_dim_' + str(ndim) + '.txt')
    rotation_matrix = np.random.default_rng(seed).standard_normal(size=(ndim, ndim))
    for i in range(ndim):
        for j in range(i):
            rotation_matrix[:, i] -= np.dot(rotation_matrix[:, i], rotation_matrix[:, j]) * rotation_matrix[:, j]
        rotation_matrix[:, i] /= np.linalg.norm(rotation_matrix[:, i])
    np.savetxt(data_path, rotation_matrix)
    return rotation_matrix


def load_rotation_matrix(func, x, rotation_matrix=None):
    """Load the rotation matrix which needs to be generated in advance.
        When `None`, the rotation matrix should have been generated and stored in txt form in advance.

    :param func: function name, a `function` object.
    :param x: decision vector, array_like of floats.
    :param rotation_matrix: rotation matrix, array_like of floats.
    :return: rotation matrix, a 2-d `ndarray` of `dtype` `np.float64`, whose shape is `(x.size, x.size)`.
    """
    x = squeeze_and_check(x)
    if rotation_matrix is None:
        if (not hasattr(func, 'pypop_rotation_matrix')) or (func.pypop_rotation_matrix.shape != (x.size, x.size)):
            data_folder = 'pypop_benchmarks_input_data'
            data_path = os.path.join(data_folder, 'rotation_matrix_' + func.__name__ + '_dim_' + str(x.size) + '.txt')
            rotation_matrix = np.loadtxt(data_path)
            if rotation_matrix.size == 1:
                rotation_matrix = np.array([[float(rotation_matrix)]])
            func.pypop_rotation_matrix = rotation_matrix
        rotation_matrix = func.pypop_rotation_matrix
    if rotation_matrix.shape != (x.size, x.size):
        raise TypeError(f'rotation matrix should have shape: {(x.size, x.size)}.')
    return rotation_matrix


def sphere(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(sphere, x, rotation_matrix)
    y = base_functions.sphere(np.dot(rotation_matrix, x))
    return y


class Sphere(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'sphere'

    def __call__(self, x, rotation_matrix=None):
        return sphere(x, rotation_matrix)


def cigar(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(cigar, x, rotation_matrix)
    y = base_functions.cigar(np.dot(rotation_matrix, x))
    return y


class Cigar(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'cigar'

    def __call__(self, x, rotation_matrix=None):
        return cigar(x, rotation_matrix)


def discus(x, rotation_matrix=None):  # also called tablet
    rotation_matrix = load_rotation_matrix(discus, x, rotation_matrix)
    y = base_functions.discus(np.dot(rotation_matrix, x))
    return y


class Discus(BaseFunction):  # also called Tablet
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'discus'

    def __call__(self, x, rotation_matrix=None):
        return discus(x, rotation_matrix)


def cigar_discus(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(cigar_discus, x, rotation_matrix)
    y = base_functions.cigar_discus(np.dot(rotation_matrix, x))
    return y


class CigarDiscus(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'cigar_discus'

    def __call__(self, x, rotation_matrix=None):
        return cigar_discus(x, rotation_matrix)


def ellipsoid(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(ellipsoid, x, rotation_matrix)
    y = base_functions.ellipsoid(np.dot(rotation_matrix, x))
    return y


class Ellipsoid(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'ellipsoid'

    def __call__(self, x, rotation_matrix=None):
        return ellipsoid(x, rotation_matrix)


def different_powers(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(different_powers, x, rotation_matrix)
    y = base_functions.different_powers(np.dot(rotation_matrix, x))
    return y


class DifferentPowers(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'different_powers'

    def __call__(self, x, rotation_matrix=None):
        return different_powers(x, rotation_matrix)


def schwefel221(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(schwefel221, x, rotation_matrix)
    y = base_functions.schwefel221(np.dot(rotation_matrix, x))
    return y


class Schwefel221(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel221'

    def __call__(self, x, rotation_matrix=None):
        return schwefel221(x, rotation_matrix)


def step(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(step, x, rotation_matrix)
    y = base_functions.step(np.dot(rotation_matrix, x))
    return y


class Step(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'step'

    def __call__(self, x, rotation_matrix=None):
        return step(x, rotation_matrix)


def schwefel222(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(schwefel222, x, rotation_matrix)
    y = base_functions.schwefel222(np.dot(rotation_matrix, x))
    return y


class Schwefel222(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel222'

    def __call__(self, x, rotation_matrix=None):
        return schwefel222(x, rotation_matrix)


def rosenbrock(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(rosenbrock, x, rotation_matrix)
    y = base_functions.rosenbrock(np.dot(rotation_matrix, x))
    return y


class Rosenbrock(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'rosenbrock'

    def __call__(self, x, rotation_matrix=None):
        return rosenbrock(x, rotation_matrix)


def schwefel12(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(schwefel12, x, rotation_matrix)
    y = base_functions.schwefel12(np.dot(rotation_matrix, x))
    return y


class Schwefel12(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel12'

    def __call__(self, x, rotation_matrix=None):
        return schwefel12(x, rotation_matrix)


def exponential(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(exponential, x, rotation_matrix)
    y = base_functions.exponential(np.dot(rotation_matrix, x))
    return y


class Exponential(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'exponential'

    def __call__(self, x, rotation_matrix=None):
        return exponential(x, rotation_matrix)


def griewank(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(griewank, x, rotation_matrix)
    y = base_functions.griewank(np.dot(rotation_matrix, x))
    return y


class Griewank(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'griewank'

    def __call__(self, x, rotation_matrix=None):
        return griewank(x, rotation_matrix)


def bohachevsky(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(bohachevsky, x, rotation_matrix)
    y = base_functions.bohachevsky(np.dot(rotation_matrix, x))
    return y


class Bohachevsky(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'bohachevsky'

    def __call__(self, x, rotation_matrix=None):
        return bohachevsky(x, rotation_matrix)


def ackley(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(ackley, x, rotation_matrix)
    y = base_functions.ackley(np.dot(rotation_matrix, x))
    return y


class Ackley(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'ackley'

    def __call__(self, x, rotation_matrix=None):
        return ackley(x, rotation_matrix)


def rastrigin(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(rastrigin, x, rotation_matrix)
    y = base_functions.rastrigin(np.dot(rotation_matrix, x))
    return y


class Rastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'rastrigin'

    def __call__(self, x, rotation_matrix=None):
        return rastrigin(x, rotation_matrix)


def scaled_rastrigin(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(scaled_rastrigin, x, rotation_matrix)
    y = base_functions.scaled_rastrigin(np.dot(rotation_matrix, x))
    return y


class ScaledRastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'scaled_rastrigin'

    def __call__(self, x, rotation_matrix=None):
        return scaled_rastrigin(x, rotation_matrix)


def skew_rastrigin(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(skew_rastrigin, x, rotation_matrix)
    y = base_functions.skew_rastrigin(np.dot(rotation_matrix, x))
    return y


class SkewRastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'skew_rastrigin'

    def __call__(self, x, rotation_matrix=None):
        return skew_rastrigin(x, rotation_matrix)


def levy_montalvo(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(levy_montalvo, x, rotation_matrix)
    y = base_functions.levy_montalvo(np.dot(rotation_matrix, x))
    return y


class LevyMontalvo(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'levy_montalvo'

    def __call__(self, x, rotation_matrix=None):
        return levy_montalvo(x, rotation_matrix)


def michalewicz(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(michalewicz, x, rotation_matrix)
    y = base_functions.michalewicz(np.dot(rotation_matrix, x))
    return y


class Michalewicz(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'michalewicz'

    def __call__(self, x, rotation_matrix=None):
        return michalewicz(x, rotation_matrix)


def salomon(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(salomon, x, rotation_matrix)
    y = base_functions.salomon(np.dot(rotation_matrix, x))
    return y


class Salomon(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'salomon'

    def __call__(self, x, rotation_matrix=None):
        return salomon(x, rotation_matrix)


def shubert(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(shubert, x, rotation_matrix)
    y = base_functions.shubert(np.dot(rotation_matrix, x))
    return y


class Shubert(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'shubert'

    def __call__(self, x, rotation_matrix=None):
        return shubert(x, rotation_matrix)


def schaffer(x, rotation_matrix=None):
    rotation_matrix = load_rotation_matrix(schaffer, x, rotation_matrix)
    y = base_functions.schaffer(np.dot(rotation_matrix, x))
    return y


class Schaffer(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schaffer'

    def __call__(self, x, rotation_matrix=None):
        return schaffer(x, rotation_matrix)
